function [MIh]=LSMI(x,y,y_type,sigma_list,lambda_list,b,fold)
%
% Least-Squares Mutual Information (with likelihood cross validation)
%
% Estimating a squared-loss variant of mutual information 
%    \frac{1}{2}\int\int (\frac{p_{xy}(x,y)}{p_x(x)p_y(y)}-1)^2 p_x(x)p_y(y) dx dy
% from input-output samples
%    { (x_i,y_i) | x_i\in R^{dx}, y_i\in R^{dy} }_{i=1}^n 
% drawn independently from a joint density p_{xy}(x,y).
% p_x(x) and p_y(y) are marginal densities of x and y, respectively.
% 
% Usage:
%       [MIh]=LSMI(x,y,y_type,sigma_list,lambda_list,b)
%
% Input:
%    x          : dx by n input sample matrix
%    y          : dy by n output sample matrix
%    y_type     : if y_type=1, delta kernel is used for y; 
%                 otherwise (or empty) Gaussian kernel is used.
%    sigma_list : (candidates of) Gaussian width
%                 If sigma_list is a vector, one of them is selected by cross validation.
%                 If sigma_list is a scalar, this value is used without cross validation
%                 If sigma_list is empty/undefined, Gaussian width is chosen from
%                 some default canditate list by cross validation
%    lambda_list: (OPTIONAL) regularization parameter
%                 If lambda_list is a vector, one of them is selected by cross validation.
%                 If lambda_list is a scalar, this value is used without cross validation
%                 If lambda_list is empty, Gaussian width is chosen from
%                 some default canditate list by cross validation
%    b          : number of Gaussian centers (if empty, b=200 is used);
%
% Output:
%    MIh        : estimated mutual information between x and y
%    score_cv   : cross validation score
%
% (c) Taiji Suzuki, Department of Mathematical Informatics, The University of Tokyo, Japan. 
%     Masashi Sugiyama, Department of Compter Science, Tokyo Institute of Technology, Japan.
%     s-taiji@stat.t.u-tokyo.ac.jp
%     sugi@cs.titech.ac.jp,

if nargin<2
  error('number of input arguments is not enough!!!')
end

n =size(x,2);
ny=size(y,2);
if n ~= ny
  error('x and y must have the same number of samples!!!')
end;

if nargin<3
  y_type=0;
end

if nargin < 4 || isempty(sigma_list)
  sigma_list=logspace(-2,2,9);
end
if nargin < 5 || isempty(lambda_list)
  lambda_list=logspace(-3,1,9);
end

if nargin<6 || isempty(b)
  b = 200;
end

if nargin<7 || isempty(fold)
  fold=5;
end


b=min(n,b);

%Gaussian centers are randomly chosen from samples
if b<n
  rand_index=randperm(n);
  u=x(:,rand_index(1:b));
  v=y(:,rand_index(1:b));
else
  u=x;
  v=y;
end

dist2_x=dist2(x,u)';
if y_type~=1
  dist2_y=dist2(y,v)';
else
  Phiy=zeros(length(v),length(y));
  for yy=unique(y)
    Phiy((v==yy),(y==yy))=1;
  end
  Phiy=[ones(1,n); Phiy];
end

if length(sigma_list)==1 && length(lambda_list)==1
  sigma_chosen=sigma_list;
  lambda_chosen=lambda_list;
  score_cv=-inf;
else
  %%%%%%%%%%%%%%%% Searching Gaussian kernel width `sigma_chosen'
  %%%%%%%%%%%%%%%% and regularization parameter `lambda_chosen' 
  fold_index=[1:fold];
  cv_index=randperm(n);
  cv_split=floor([0:n-1]*fold./n)+1;
  scores_cv=zeros(length(sigma_list),length(lambda_list));

  for sigma_index=1:length(sigma_list)
    sigma=sigma_list(sigma_index);
    Phix_sigma=[ones(1,n); exp(-dist2_x/(2*sigma^2))];
    if y_type~=1
      Phiy_sigma=[ones(1,n); exp(-dist2_y/(2*sigma^2))];
    else
      Phiy_sigma=Phiy;
    end
    Phi_sigma=Phix_sigma.*Phiy_sigma;

    for i=fold_index
      cv_index_tmp=cv_index(cv_split==i);
      HH_cv_Phix(:,:,i)=Phix_sigma(:,cv_index_tmp)*Phix_sigma(:,cv_index_tmp)';
      HH_cv_Phiy(:,:,i)=Phiy_sigma(:,cv_index_tmp)*Phiy_sigma(:,cv_index_tmp)';
      hh_cv(:,:,i)=mean(Phi_sigma(:,cv_index_tmp),2);
      n_cv(i)=length(cv_index_tmp);
    end

    for i=fold_index
      cv_index_tr=fold_index(fold_index~=i);
      Hh_cv_tr=sum(HH_cv_Phix(:,:,cv_index_tr),3) ...
               .*sum(HH_cv_Phiy(:,:,cv_index_tr),3)/(sum(n_cv(cv_index_tr))^2);
      Hh_cv_te=HH_cv_Phix(:,:,i).*HH_cv_Phiy(:,:,i)/(n_cv(i)^2);
      hh_cv_tr=mean(hh_cv(:,:,cv_index_tr),3);
      hh_cv_te=hh_cv(:,:,i);
      for lambda_index=1:length(lambda_list)
        lambda=lambda_list(lambda_index);
        alphah_cv=(Hh_cv_tr+lambda*eye(b+1))\hh_cv_tr;
        wh_cv=alphah_cv'*Hh_cv_te*alphah_cv/2-hh_cv_te'*alphah_cv;
        scores_cv(sigma_index,lambda_index)=scores_cv(sigma_index,lambda_index)+wh_cv/fold;
      end % fold
    end % lambda
  end % sigma
  [scores_cv_tmp,lambda_chosen_index]=min(scores_cv,[],2);
  [score_cv,sigma_chosen_index]=min(scores_cv_tmp);
  lambda_chosen=lambda_list(lambda_chosen_index(sigma_chosen_index));
  sigma_chosen=sigma_list(sigma_chosen_index);
end %length(sigma_list)==1 && length(lambda_list)==1

%%%%%%%%%%%%%%%% Computing the final solution `MIh'
Phix=[ones(1,n); exp(-dist2_x/(2*sigma_chosen^2))];
if y_type~=1
  Phiy=[ones(1,n); exp(-dist2_y/(2*sigma_chosen^2))];
end
Phi=Phix.*Phiy;
Hh=(Phix*Phix').*(Phiy*Phiy')/(n^2);
hh=mean(Phi,2);
alphah=(Hh+lambda_chosen*eye(b+1))\hh;
MIh=hh'*alphah-alphah'*Hh*alphah/2-1/2;

